# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.


from metakernel import Magic, option
import urllib
try:
    import urlparse
except ImportError:
    from urllib import parse as urlparse
import os

try:
    urllib.URLopener
    def download(url, filename):
        opener = urllib.URLopener()
        opener.retrieve(url, filename)
except: # python3
    import urllib.request
    def download(url, filename):
        g = urllib.request.urlopen(url)
        with open(filename, 'wb') as f:
            f.write(g.read())        

class DownloadMagic(Magic):

    @option(
        '-f', '--filename', action='store', default=None,
        help='use the provided name as filename'
    )
    def line_download(self, url, filename=None):
        """
        %download URL [-f FILENAME] - download file from URL

        This line magic will download and save a file. By
        default it will use the same filename as the URL.
        You can give it another name using -f.

        Examples:
            %%download http://some/file/from/internet.txt -f myfile.txt
            %%download http://some/file/from/program.ss

        """
        if filename is None:
            if url.count(" ") == 1:
                url, filename = url.split(" ")
            elif url.count(" ") == 0:
                parts = urlparse.urlsplit(url)
                #('http', 'example.com', '/somefile.zip', '', '')
                path = parts[2]
                filename = os.path.basename(path)
            else:
                raise Exception("invalid arguments to %%download: '%s'" % url)
        basename, extname = os.path.splitext(filename)
        if extname == '':
            filename += ".html"
        filename = filename.replace("~", "")
        filename = filename.replace("%20", "_")
        try:
            download(url, filename)
            self.kernel.Print("Downloaded '%s'." % filename)
        except Exception as e:
            self.kernel.Error(str(e))


def register_magics(kernel):
    kernel.register_magics(DownloadMagic)

def register_ipython_magics():
    from metakernel import IPythonKernel
    from IPython.core.magic import register_line_magic
    kernel = IPythonKernel()
    magic = DownloadMagic(kernel)
    # Make magics callable:
    kernel.line_magics["download"] = magic

    @register_line_magic
    def download(line):
        kernel.call_magic("%download " + line)



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from __future__ import print_function
from metakernel import Magic, pexpect
from metakernel.replwrap import cmd, bash
import os


class ShellMagic(Magic):

    def __init__(self, kernel):
        super(ShellMagic, self).__init__(kernel)
        self.repl = None
        self.cmd = None
        self.start_process()

    def line_shell(self, *args):
        """
        %shell COMMAND - run the line as a shell command

        This line command will run the COMMAND in the bash shell.

        Examples:
            %shell ls -al
            %shell cd

         Note: this is a persistent connection to a shell.
         The working directory is synchronized to that of the notebook
         before and after each call.

        You can also use "!" instead of "%shell".

        """
        # get in sync with the cwd
        self.eval('cd "%s"' % os.getcwd().replace(os.path.sep, '/'))

        command = " ".join(args)
        self.eval(command, True)

        if self.cmd == 'cmd':
            cwd = self.eval('echo %cd%')
        else:
            cwd = self.eval('pwd')

        if os.path.exists(cwd):
            os.chdir(cwd)

    def eval(self, cmd, incremental=False):
        stream_handler = self.kernel.Print if incremental else None
        return self.repl.run_command(cmd, timeout=None,
                                     stream_handler=stream_handler)

    def start_process(self):
        if self.repl is not None:
            self.repl.child.terminate()

        if not self.cmd:
            if os.name == 'nt':
                self.cmd = 'cmd'
                self.repl = cmd()
            elif pexpect.which('bash'):
                self.cmd = 'bash'
                self.repl = bash()
            elif pexpect.which('sh'):
                self.cmd = 'sh'
                self.repl = bash(command='sh')
            else:
                msg = "The command was not found or was not executable: sh"
                raise Exception(msg)

    def cell_shell(self):
        """
        %%shell - run the contents of the cell as shell commands

        This shell command will run the cell contents in the bash shell.

        Example:
            %%shell
               cd ..
               ls -al

       Note: this is a persistent connection to a shell.
         The working directory is synchronized to that of the notebook
         before and after each call.

        You can also use "!!" instead of "%%shell".
        """
        self.line_shell(self.code)
        self.evaluate = False

    def get_completions(self, info):
        if self.cmd == 'cmd':
            return []
        command = 'compgen -cdfa "%s"' % info['code']
        completion_text = self.eval(command)
        return completion_text.split()

    def get_help_on(self, info, level=0):
        expr = info['code'].rstrip()
        if self.cmd == 'cmd':
            resp = self.eval('help %s' % expr)
        elif level == 0:
            resp = self.eval('%s --help' % expr)
        else:
            resp = self.eval('man %s' % expr)
        if resp and not ': command not found' in resp:
            return resp
        else:
            return "Sorry, no help is available on '%s'." % expr


def register_magics(kernel):
    kernel.register_magics(ShellMagic)



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.


from metakernel import Magic
from IPython.display import HTML

class DotMagic(Magic):

    def line_dot(self, code):
        """
        %dot CODE - render code as Graphviz image

        This line magic will render the Graphiz CODE, and render 
        it as an image.

        Example:
            %dot graph A { a->b };

        """
        try:
            import pydot
        except:
            raise Exception("You need to install pydot")
        graph = pydot.graph_from_dot_data(str(code))
        svg = graph.create_svg()
        if hasattr(svg, "decode"):
            svg = svg.decode("utf-8")
        html = HTML(svg)
        self.kernel.Display(html)

    def cell_dot(self):
        """
        %%dot - render contents of cell as Graphviz image

        This cell magic will send the cell to the browser as
        HTML.

        Example:
            %%dot

            graph A { a->b };
        """
        try:
            import pydot
        except:
            raise Exception("You need to install pydot")
        graph = pydot.graph_from_dot_data(str(self.code))
        svg = graph.create_svg()
        if hasattr(svg, "decode"):
            svg = svg.decode("utf-8")
        html = HTML(svg)
        self.kernel.Display(html)
        self.evaluate = False

def register_magics(kernel):
    kernel.register_magics(DotMagic)

def register_ipython_magics():
    from metakernel import IPythonKernel
    from IPython.core.magic import register_cell_magic
    kernel = IPythonKernel()
    magic = DotMagic(kernel)

    @register_cell_magic
    def dot(line, cell):
        """
        %%dot - evaluate cell contents as a dot diagram.
        """
        magic.code = cell
        magic.cell_dot()




import os
import sys
from metakernel.tests.utils import (get_kernel, get_log_text,
                                    clear_log_text, EvalKernel)
PY3 = (sys.version_info[0] >= 3)

def test_set_get_int_magic():
    kernel = get_kernel(EvalKernel)
    kernel.do_execute("%set x 42")
    kernel.do_execute("%get x")
    text = get_log_text(kernel)
    assert "42" in text, text

def test_set_get_list_magic():
    kernel = get_kernel(EvalKernel)
    kernel.do_execute("%set variable [1., 2., 3., 4.]")
    kernel.do_execute("%get variable")
    text = get_log_text(kernel)
    assert "[1.0, 2.0, 3.0, 4.0]" in text, text

def test_set_get_range_magic():
    kernel = get_kernel(EvalKernel)
    kernel.do_execute("%set variable range(2)")
    kernel.do_execute("%get variable")
    text = get_log_text(kernel)
    if PY3:
        assert "range(0, 2)" in text, text
    else:
        assert "[0, 1]" in text, text




from metakernel.tests.utils import get_kernel


def test_load_magic():
    kernel = get_kernel()
    ret = kernel.do_execute("%%load %s" % __file__)
    assert 'def test_load_magic' in ret['payload'][0]['text']




from metakernel import Parser
import os


def test_parser():
    p = Parser()

    info = p.parse_code('import nump')
    assert info['help_obj'] == info['obj'] == 'nump'

    assert p.parse_code('%python impor', 0, 10)['magic']['name'] == 'python'
    assert p.parse_code('oct(a,')['help_obj'] == 'oct'
    assert p.parse_code('! ls')['magic']['name'] == 'shell'

    info = p.parse_code('%help %lsmagic', 0, 10)
    assert info['help_obj'] == '%lsmagic'
    assert info['obj'] == '%lsm'

    info = p.parse_code('%%python\nprint("hello, world!",')
    assert info['help_obj'] == 'print'

    info = p.parse_code('%lsmagic')
    assert info['help_obj'] == '%lsmagic'

    info = p.parse_code('%')
    assert info['magic']['type'] == 'line'


def test_scheme_parser():
    function_call_regex = r'\(([^\d\W][\w\.]*)[^\)\()]*\Z'
    p = Parser(function_call_regex=function_call_regex)

    info = p.parse_code('(oct a b ')
    assert info['help_obj'] == 'oct'


def test_path_completions():
    p = Parser()

    if not os.name == 'nt':
        code = '/usr/bi'
        assert 'bin/' in p.parse_code(code)['path_matches']
    code = '~/.bashr'
    assert 'bashrc' in p.parse_code(code)['path_matches']

    for f in os.listdir('.'):
        if f.startswith('.'):
            if os.path.isdir(f):
                assert f[1:] + os.sep in p.parse_code('.')['path_matches'], f
            else:
                assert f[1:] in p.parse_code('.')['path_matches']


def test_complete0():
    p = Parser()
    info = p.parse_code('abcdefghijklmnop', 0, 4)
    assert info['obj'] == 'abcd', info


def get_parser():
    p = Parser()
    try:
        os.mkdir("/tmp/Test Dir")
    except OSError:
        pass  # dir exists
    open("/tmp/Test Dir/test.txt", "w").close()
    return p


def test_complete1():
    p = get_parser()
    info = p.parse_code('/tmp/')
    assert "Test\ Dir/" in info['path_matches'], info['path_matches']


def test_complete2():
    p = get_parser()
    info = p.parse_code('open("/tmp/')
    assert "Test Dir/" in info['path_matches'], info


def test_complete3():
    p = get_parser()
    info = p.parse_code('/tmp/Test Dir/temp.txt', 0, 14)
    assert "test.txt" in info['path_matches'], info


def test_complete4():
    p = get_parser()
    info = p.parse_code('/tmp/Test Dir')
    assert 'Dir/test.txt' in info['path_matches'], info


def test_complete5():
    p = get_parser()
    info = p.parse_code('/tmp/Test Dir/')
    assert "test.txt" in info['path_matches'], info


def test_complete6():
    p = get_parser()
    info = p.parse_code('/tmp/Test')
    assert "Test\ Dir/" in info['path_matches'], info['path_matches']


def test_complete7():
    p = get_parser()
    info = p.parse_code('/tmp/Test Dir/test.txt')
    assert not info['path_matches'], info


def test_complete8():
    p = get_parser()
    info = p.parse_code('/tmp/Test Dir/', 0, 9)
    assert 'Test\ Dir/' in info['path_matches'], info


def test_complete9():
    p = get_parser()
    info = p.parse_code('fluff\n/tmp/Test ')
    assert 'Dir/' in info['path_matches'], info


def test_complete10():
    p = get_parser()
    info = p.parse_code('/tmp/Test\\ Dir')
    assert 'Dir/test.txt' in info['path_matches']



from metakernel.tests.utils import get_kernel, EvalKernel, get_log_text


def test_restart_magic():
    kernel = get_kernel(EvalKernel)

    kernel.do_execute('a=1')
    kernel.do_execute("%restart")
    kernel.do_execute('a')

    text = get_log_text(kernel)
    assert 'NameError' in text



# Convenience imports from pexpect
from __future__ import absolute_import
import os
import signal
from pexpect import is_executable_file, EOF, TIMEOUT, __file__ as PEXPECT_DIR

try:
    from pexpect import spawn as pty_spawn
    import pty
except ImportError:
    from pexpect.popen_spawn import PopenSpawn
    pty = None


def spawn(command, args=[], timeout=30, maxread=2000,
          searchwindowsize=None, logfile=None, cwd=None, env=None,
          ignore_sighup=True, echo=True, encoding='utf-8', **kwargs):
    '''This is the main entry point for Pexpect. Use this functio to start
    and control child applications.

    See https://github.com/pexpect/pexpect/blob/master/pexpect/pty_spawn.py
    for more information.
    '''
    codec_errors = kwargs.get('codec_errors', kwargs.get('errors', 'strict'))
    if pty is None:
        if args:
            command += ' ' + ' '.join(args)
        child = PopenSpawn(command, timeout=timeout, maxread=maxread,
                           searchwindowsize=searchwindowsize,
                           logfile=logfile, cwd=cwd, env=env,
                           encoding=encoding, codec_errors=codec_errors)
        child.echo = echo
    else:
        try:
            # Signal handlers are inherited by forked processes, and we can't easily
            # reset it from the subprocess. Since kernelapp ignores SIGINT except in
            # message handlers, we need to temporarily reset the SIGINT handler here
            # so that the child and its children are interruptible.
            try:
                sig = signal.signal(signal.SIGINT, signal.SIG_DFL)
            except ValueError:
                # Only Main Thread can handle signals
                sig = None
            child = pty_spawn(command, args=args, timeout=timeout,
                              maxread=maxread,
                              searchwindowsize=searchwindowsize,
                              logfile=logfile, cwd=cwd, env=env,
                              encoding=encoding, codec_errors=codec_errors)
        finally:
            if sig:
                signal.signal(signal.SIGINT, sig)
    return child


# For backwards compatibility
spawnu = spawn


def which(filename):
    '''This takes a given filename; tries to find it in the environment path;
    then checks if it is executable. This returns the full path to the filename
    if found and executable. Otherwise this returns None.'''

    # Special case where filename contains an explicit path.
    if os.path.dirname(filename) != '' and is_executable_file(filename):
        return filename
    if 'PATH' not in os.environ or os.environ['PATH'] == '':
        p = os.defpath
    else:
        p = os.environ['PATH']
    pathlist = p.split(os.pathsep)
    for path in pathlist:
        ff = os.path.join(path, filename)
        if pty:
            if is_executable_file(ff):
                return ff
        else:
            pathext = os.environ.get('Pathext', '.exe;.com;.bat;.cmd')
            pathext = pathext.split(os.pathsep) + ['']
            for ext in pathext:
                if os.access(ff + ext, os.X_OK):
                    return ff + ext
    return None



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.


from metakernel import Magic
import os


class CDMagic(Magic):

    def line_cd(self, path='.'):
        """
        %cd PATH - change current directory of session

        This line magic is used to change the directory of the
        notebook or console.

        Note that this is not the same directory as used by
        the %shell magics.

        Example:
            %cd ..
        """
        path = os.path.expanduser(path)
        try:
            os.chdir(path)
            retval = os.path.abspath(path)
        except Exception as e:
            self.kernel.Error(str(e))
            retval = None
        if retval:
            self.kernel.Print(retval)


def register_magics(kernel):
    kernel.register_magics(CDMagic)



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic, option


class PlotMagic(Magic):

    @option(
        '-s', '--size', action='store',
        help='Pixel size of plots, "width,height"'
    )
    @option(
        '-f', '--format', action='store',
        help='Plot format (png, svg or jpg).'
    )
    @option(
        '-b', '--backend', action='store', default='inline',
        help='Backend selection'
    )
    @option(
        '-r', '--resolution', action='store',
        help='Resolution in pixels per inch'
    )
    @option(
        '-w', '--width', action='store',
        help='Plot width in pixels'
    )
    @option(
        '-h', '--height', action='store',
        help='Plot height in pixels'
    )
    def line_plot(self, *args, **kwargs):
        """
        %plot [options] backend - configure plotting for the session.

        This line magic will configure the plot settings for this
        language.

        Examples:
            %plot qt --format=png
            %plot inline -w 640

        Note: not all languages may support the %plot magic, and not all
        options may be supported.
        """
        if args and not args[0].startswith('-'):
            kwargs['backend'] = args[0]
        if 'size' in kwargs and kwargs['size'] is not None:
            width, height = kwargs['size']
            kwargs['width'] = int(width)
            kwargs['height'] = int(height)
        # Remove empty options so ".setdefault" will work.
        for key in ['resolution', 'format', 'size', 'width', 'height']:
            if key in kwargs and kwargs[key] is None:
                del kwargs[key]
        self.kernel.plot_settings = kwargs
        self.kernel.handle_plot_settings()


def register_magics(kernel):
    kernel.register_magics(PlotMagic)



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.


from metakernel import Magic


class HelpMagic(Magic):

    def help_strings(self):
        suffixes = [
            "item{0}{0} - get detailed, technical information on item",
            "item{0}  - get help on item",
        ]
        prefixes = [
            "{0}{0}item - get detailed, technical information on item",
            "{0}item  - get help on item",
        ]
        strings = []
        if self.kernel.help_suffix:
            strings += [s.format(self.kernel.help_suffix['help'])
                        for s in suffixes]
        if 'help' in self.kernel.magic_prefixes:
            strings += [p.format(self.kernel.magic_prefixes['help'])
                        for p in prefixes]
        return sorted(strings)

    def line_help(self, text):
        """
        %help TEXT - get help on the given text

        This line magic shows the help for TEXT. Shows the help in the
        system pager that opens at bottom of screen.

        Example:
            %help dir

        """
        return self.get_help_on(text, 0, False)

    def cell_help(self, text):
        """
        %%help TEXT - get detailed help on the given text

        This line magic looks like a cell magic, but it really gets
        more detailed help on TEXT. Shows the help in the system
        pager that opens at bottom of screen.

        Example:
           %%help dir

        """
        return self.get_help_on(text, 1, False)

    def get_help_on(self, text, level, none_on_fail=False, cursor_pos=-1):
        """
        Examples
        ========

        All of the following give help on np.ones using the %python magic:

        %help %python np.ones
        %python np.ones?

        To get help on a magic itself, we would write one of the following:

        %python?
        %help %python
        ?%python
        """
        text = self._prep_text(text)
        if not text:
            return self.kernel.get_usage()

        info = self.kernel.parse_code(text, cursor_end=cursor_pos)

        if info['magic']:

            minfo = info['magic']
            errmsg = "No such %s magic '%s'" % (minfo['type'], minfo['name'])

            if minfo['type'] == 'line':
                magic = self.kernel.line_magics.get(minfo['name'], None)
                if minfo['args']:
                    info = self.kernel.parse_code(minfo['args'])
                elif magic:
                    return magic.get_help(minfo['type'], minfo['name'],
                                          level)

            else:
                magic = self.kernel.cell_magics.get(minfo['name'], None)

                if minfo['code']:
                    info = self.kernel.parse_code(minfo['code'])
                elif minfo['args']:
                    info = self.kernel.parse_code(minfo['args'])
                elif magic:
                    return magic.get_help(minfo['type'], minfo['name'],
                                          level)
                else:
                    return ("No such %s magic named '%s', so can't really help with that" % 
                            (minfo["type"], minfo["name"]))
            if magic:
                the_help = magic.get_help_on(info, level)
                self.kernel.log.error('got')
                self.kernel.log.error(the_help)
                return the_help

            elif not info['magic']['name']:
                return self.kernel.get_usage()

            else:
                return errmsg

        else:
            return self.kernel.get_kernel_help_on(info, level, none_on_fail)

    def _prep_text(self, text):
        text = text.strip()

        magic = self.kernel.magic_prefixes['magic']
        prefix = self.kernel.magic_prefixes['help']
        suffix = self.kernel.help_suffix

        text = text.replace('{0}{0}help'.format(magic), '')
        text = text.replace('{0}help'.format(magic), '')

        if prefix:
            while text.startswith(prefix):
                text = text[1:]

        if suffix:
            while text.endswith(suffix):
                text = text[:-1]

        return text


def register_magics(kernel):
    kernel.register_magics(HelpMagic)




from metakernel.tests.utils import get_kernel, get_log_text, EvalKernel
import os
import time

def test_parallel_magic():
    kernel = get_kernel(EvalKernel)
    # start up an EvalKernel on each node:
    kernel.do_execute("%parallel metakernel_python MetaKernelPython", False)
    # Now, execute something on each one:
    kernel.do_execute("%px cluster_rank", False)
    results = get_log_text(kernel)
    assert "[0, 1, 2]" in results, results
    
# Starting the cluster from here doesn't work with nosetests
# so we start `ipcluster` before we test.

#def setup_func():
#    ## start up a cluster in the background with three nodes:
#    os.system("ipcluster start --n=3 &")

#def teardown():
#    ## shutdown the cluster:
#    os.system("ipcluster stop")




# -*- coding: utf-8 -*-

# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.


from metakernel import Magic
from IPython.display import HTML, Javascript

class DebugMagic(Magic):

    def cell_debug(self, dummy):
        """
        %%debug - step through the code expression by expression

        This cell magic will step through the code in the cell,
        if the kernel supports debugging.

        Example:
            %%debug

            (define x 1)
        """

        html_code = """
<style type="text/css">
      .breakpoints {width: 1.5em;}
      .styled-background { background-color: #ff7; }
</style>

<div class="animation" align="left" width="400">
  <button onclick="step()">Step</button>
  <button onclick="play()">Play</button>
  <button onclick="pause()">Pause</button>
  <button onclick="stop()">Stop</button>
  <button onclick="clear_breakpoints()">Clear Breakpoints</button>
  <button onclick="reset()">Reset</button><br/>
  Speed: <input type="range" id="speed" min="1" max="100"/>
  Inspect: <input type="text" id="inspect" onkeydown="if (event.keyCode == 13) inspector()"/>
  <table width="100%">
  <tr>
    <td width="50%" style="vertical-align: top"><div id="result_stream"/></td>
    <td width="50%" style="vertical-align: top"><div id="result_output"/></td>
  </tr>
  </table>
</div>

<script>

function inspector() {
    var v = document.getElementById("inspect").value;
    document.getElementById("inspect").value = "";
    var msg_id = kernel.execute("~~META~~: inspect " + v,
                                callbacks,
                                {silent: false});
}

function clear_breakpoints() {
    for (var n = 0 ; n < cell.code_mirror.doc.size; n++) {
        cell.code_mirror.setGutterMarker(n, "breakpoints", null);
    }
}

function makeMarker() {
    var marker = document.createElement("div");
    marker.style.color = "#822";
    marker.innerHTML = "●";
    return marker;
}

function breakpoint_q(line) {
    var info = cell.code_mirror.lineInfo(line - 1);
    return (info && info.gutterMarkers != null);
}

var cell = IPython.notebook.get_selected_cell();
if (cell) {
    // The following can be replaced once get_prev_cell is fixed
    // see https://github.com/ipython/ipython/pull/6565
    var index = IPython.notebook.find_cell_index(cell);
    cell = IPython.notebook.get_cell(index - 1);
    cell.code_mirror.setOption("lineNumbers", true);
    if (cell.code_mirror.getOption("gutters").length < 2) {
        cell.code_mirror.setOption("gutters", ["CodeMirror-linenumbers", "breakpoints"]);
        cell.code_mirror.refresh();
        cell.code_mirror.on("gutterClick", function(cm, n) {
          var info = cm.lineInfo(n);
          cm.setGutterMarker(n, "breakpoints", info.gutterMarkers ? null : makeMarker());
        });
    }
}

var mt;
var i;

function highlight(cell, data) {
    if (cell) {
        var line1 = data[0];
        var col1 = data[1];
        var line2 = data[2];
        var col2 = data[3];
        if (mt) {
            mt.clear();
        }
        mt = cell.code_mirror.markText({line: line1 - 1, ch: col1 - 1},
                                       {line: line2 - 1, ch: col2 },
                                       {className: "styled-background"});
    }
}

var running = false;  // is it running?
var timer = null;
var kernel = IPython.notebook.kernel;

function handle_output(out){
    var res = null;
    var data = null;
    if (out.msg_type == "stream") {
        res = out.content.text;
        document.getElementById("result_stream").innerText = res.toString() + document.getElementById("result_stream").innerText;
    } else if (out.msg_type === "pyout") {
        // if output is a python object
        res = out.content.data["text/plain"];
        document.getElementById("result_stream").innerText = res.toString() + document.getElementById("result_stream").innerText;
    } else if (out.msg_type == "pyerr") {
        // if output is a python error
        res = out.content.data["text/plain"];
        document.getElementById("result_stream").innerText = res.toString() + document.getElementById("result_stream").innerText;
    } else if (out.msg_type == "execute_result") {
        var str = out.content.data["text/plain"];
        if (str.indexOf("\\"highlight: ") >= 0) { // is a highlight response:
            data = JSON.parse(str.substring(12, str.length - 1));
            var speed = Number(document.getElementById("speed").value);
            if (data) {
                var breakp = breakpoint_q(data[0]);
                if (speed < 100 || !running || breakp) { // max_speed
                    highlight(cell, data);
                }
            }
            if (running) {
                if (data && breakpoint_q(data[0])) {
                    pause();
                } else {
                    timer = setTimeout(step, 2000/speed);
                }
            }
        } else { // display result
            res = str;
            if (res.indexOf("u") == 0)
                res = res.substring(2, res.length - 1) + "\\n";
            if (res) {
                document.getElementById("result_output").innerText = res.toString() + document.getElementById("result_output").innerText;
            }
        }
    } else {
        // if output is something we haven't thought of
        res = out.toString();
        document.getElementById("result_stream").innerText = res.toString() + document.getElementById("result_stream").innerText;
    }
}

var callbacks = { 'iopub' : {'output' : handle_output}};

function pause() {
    running = false;
    document.getElementById("result_output").innerText = "Pause\\n" + document.getElementById("result_output").innerText;
    if (timer) {
        clearInterval(timer);
        timer = null;
    }
}

function play() {
    pause();
    running = true;
    step();
}

function step() {
    var msg_id = kernel.execute("~~META~~: step",
                                callbacks,
                                {silent: false});
}

function stop() {
    pause();
    if (mt) {
        mt.clear();
    }
    var msg_id = kernel.execute("~~META~~: stop",
                                callbacks,
                                {silent: false});
}

function reset() {
    stop();
    document.getElementById("result_output").innerText = "";
    document.getElementById("result_stream").innerText = "";
    var msg_id = kernel.execute("~~META~~: reset",
                                callbacks,
                                {silent: false});
}

</script>
"""
        import time
        html = HTML(html_code)
        self.kernel.Display(html)
        time.sleep(.1)
        data = self.kernel.initialize_debug("\n" + self.code) ## add a line so line numbers will be correct
        time.sleep(.1)
        if data.startswith("highlight: "):
            self.kernel.Display(Javascript("highlight(cell, %s);" % data[11:]))
        time.sleep(.1)
        self.evaluate = False

def register_magics(kernel):
    kernel.register_magics(DebugMagic)




from metakernel.tests.utils import get_kernel

EXECUTION = ""

def test_include_magic():
    global EXECUTION
    kernel = get_kernel()
    def do_execute_direct(code):
        global EXECUTION
        EXECUTION = code
    kernel.do_execute_direct = do_execute_direct
    FILE = __file__
    if FILE.endswith(".pyc"):
        FILE = FILE[:-1]
    kernel.do_execute("%%include %s" % FILE)
    assert "metakernel" in EXECUTION
    assert ("AND " + "THIS") not in EXECUTION

    EXECUTION = ""
    kernel.do_execute(("%%include %s\nAND" + " THIS") % FILE)
    assert "metakernel" in EXECUTION
    assert ("AND " + "THIS") in EXECUTION

    EXECUTION = ""
    kernel.do_execute("%%include '%s' '%s'" % (FILE, FILE))
    assert "metakernel" in EXECUTION
    assert ("AND " + "THIS") not in EXECUTION



from metakernel import MetaKernel
try:
    from jupyter_client import session as ss
except ImportError:
    from IPython.kernel.zmq import session as ss
import zmq
import logging

try:
    from StringIO import StringIO
except ImportError:
    from io import StringIO


class EvalKernel(MetaKernel):
    implementation = 'Eval'
    implementation_version = '1.0'
    language = 'python'
    language_version = '0.1'
    banner = "Eval kernel - evaluates simple Python statements and expressions"

    def set_variable(self, name, value):
        """
        Set a variable in the kernel language.
        """
        python_magic = self.line_magics['python']
        python_magic.env[name] = value

    def get_variable(self, name):
        """
        Get a variable from the kernel language.
        """
        python_magic = self.line_magics['python']
        return python_magic.env[name]

    def do_execute_direct(self, code):
        python_magic = self.line_magics['python']
        return python_magic.eval(code.strip())

    def do_execute_meta(self, code):
        if code == "reset":
            return "RESET"
        elif code == "stop":
            return "STOP"
        elif code == "step":
            return "STEP"
        elif code.startswith("inspect "):
            return "INSPECT"
        else:
            raise Exception("Unknown meta command: '%s'" % code)

    def initialize_debug(self, code):
        return "highlight: [%s, %s, %s, %s]" % (0, 0, 1, 0)


def get_kernel(kernel_class=MetaKernel):
    log = logging.getLogger('test')
    log.setLevel(logging.DEBUG)

    for hdlr in log.handlers:
        log.removeHandler(hdlr)

    hdlr = logging.StreamHandler(StringIO())
    hdlr.setLevel(logging.DEBUG)
    log.addHandler(hdlr)

    context = zmq.Context.instance()
    iopub_socket = context.socket(zmq.PUB)

    kernel = kernel_class(session=ss.Session(), iopub_socket=iopub_socket,
                          log=log)
    return kernel


def clear_log_text(kernel):
    kernel.log.handlers[0].stream.truncate(0)
    kernel.log.handlers[0].stream.seek(0)


def get_log_text(kernel):
    return kernel.log.handlers[0].stream.getvalue()




# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic, option
import os

class IncludeMagic(Magic):

    def line_include(self, filenames):
        """
        %include FILENAME ... - include code from filename into this code

        This line magic will get the contents of a file and include it
        in this cell evaluation.

        You can have multiple %include's at the beginning of a cell,
        and they will be included in order.

        Examples:
            %include myprog.py
            %include myprog1.py myprog2.py
        """
        text = ""
        filenames = filenames.split()
        prefix = self.kernel.magic_prefixes['magic']
        for filename in filenames:
            if filename.startswith("~"):
                filename = os.path.expanduser(filename)
            filename = os.path.abspath(filename)
            text += open(filename).read() + "\n"
        if self.code.lstrip().startswith(prefix):
            lines = self.code.lstrip().split("\n")
            new_lines = []
            need_to_insert = True
            for line in lines:
                if need_to_insert and not line.startswith(prefix):
                    new_lines.append(text)
                    need_to_insert = False
                new_lines.append(line)
            if need_to_insert:
                new_lines.append(text)
            self.code = "\n".join(new_lines)
        else:
            self.code = text  + self.code

def register_magics(kernel):
    kernel.register_magics(IncludeMagic)




# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic, option
import os

class RunMagic(Magic):

    @option(
        '-l', '--language', action='store', default=None,
        help='use the provided language name as kernel'
    )
    def line_run(self, filename, language=None):
        """
        %run [--language LANG] FILENAME - run code in filename by
           kernel

        This magic will take the code in FILENAME and run it. The
        exact details of how the code runs are deterimined by your
        language.

        The --language LANG option will prefix the file contents with
        "%%LANG". You may also put information in the cell which will
        appear before the contents of the file.

        Examples:
            %run filename.ss
            %run -l python filename.py

            %kx calysto_scheme.kernel CalystoScheme
            %run --language kx filename.ss
            %run --language "kx default" filename.ss

        Note: not all languages may support %run.
        """
        if filename.startswith("~"):
            filename = os.path.expanduser(filename)
        filename = os.path.abspath(filename)
        if language is None:
            self.kernel.do_execute_file(filename)
        else:
            code = "".join(open(filename).readlines())
            self.code = "%%" + language + "\n" + self.code + code

def register_magics(kernel):
    kernel.register_magics(RunMagic)




# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic
from IPython.display import Latex

class LatexMagic(Magic):

    def line_latex(self, text):
        """
        %latex TEXT - display text as LaTeX

        This line magic will display the TEXT on the line as LaTeX.

        Example:
            %latex x_1 = \dfrac{a}{b}

        """
        latex = Latex(text)
        self.kernel.Display(latex)

    def cell_latex(self):
        """
        %%latex - display contents of cell as LaTeX

        This cell magic will display the TEXT in the cell as LaTeX.

        Example:
            %%latex 
            x_1 = \dfrac{a}{b}

            x_2 = a^{n - 1}
        """
        latex = Latex(self.code)
        self.kernel.Display(latex)
        self.evaluate = False

def register_magics(kernel):
    kernel.register_magics(LatexMagic)



from metakernel import Magic, option
from IPython.display import IFrame
    
class BrainMagic(Magic):
    def cell_brain(self):
        """
        %%brain - run a cell as brain control code
        for a calysto.simulation.

        Requires calysto.

        Examples:
           %%brain
           robot.forward(1)
        """
        text = self.code
        pre_code = """
from calysto.simulation import *
robot = get_robot()
def brain():
"""
        post_code = """robot.brain = brain"""
        new_code = ("    ".join(line + "\n" for line in text.split("\n")))
        self.code = pre_code + new_code + post_code

def register_magics(kernel):
    kernel.register_magics(BrainMagic)

def register_ipython_magics():
    from metakernel import IPythonKernel
    from IPython.core.magic import register_cell_magic
    kernel = IPythonKernel()
    magic = BrainMagic(kernel)

    @register_cell_magic
    def brain(line, cell):
        from IPython import get_ipython
        ipkernel = get_ipython()
        magic.code = cell
        magic.cell_brain()
        ipkernel.kernel.do_execute(magic.code, silent=True)



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic, option

try:
    from calysto_scheme import scheme
except:
    scheme = None

class SchemeMagic(Magic):
    def __init__(self, kernel):
        super(SchemeMagic, self).__init__(kernel)
        self.retval = None

    def line_scheme(self, *args):
        """
        %scheme CODE - evaluate code as Scheme

        This line magic will evaluate the CODE (either expression or
        statement) as Scheme code.

        Examples:
            %scheme (define x 42)
            %scheme (import "math")
            %scheme (+ x + math.pi)

        """
        code = " ".join(args)
        self.retval = self.eval(code)

    def eval(self, code):
        if scheme:
            return scheme.execute_string_rm(code.strip())
        else:
            raise Exception("calysto_scheme is required")

    @option(
        "-e", "--eval_output", action="store_true", default=False,
        help="Use the retval value from the Scheme cell as code in the kernel language."
    )
    def cell_scheme(self, eval_output=False):
        """
        %%scheme - evaluate contents of cell as Scheme

        This cell magic will evaluate the cell (either expression or
        statement) as Scheme code.

        The -e or --eval_output flag signals that the retval value expression
        will be used as code for the cell to be evaluated by the host
        language.

        Examples:
            %%scheme
            (define x 42)

            %%scheme
            (import "math")
            (define retval (+ x math.pi))

            %%scheme -e
            (define retval "this = code")

            %%scheme -e
            "this = code"
        """
        if self.code.strip():
            if eval_output:
                self.code = self.eval(self.code)
                self.evaluate = True
            else:
                self.retval = self.eval(self.code)
                self.evaluate = False

    def post_process(self, retval):
        if retval is not None:
            return retval
        else:
            return self.retval

def register_magics(kernel):
    kernel.register_magics(SchemeMagic)

def register_ipython_magics():
    from metakernel import IPythonKernel
    from IPython.core.magic import register_line_magic, register_cell_magic
    kernel = IPythonKernel()
    magic = SchemeMagic(kernel)

    @register_line_magic
    def scheme(line):
        magic.line_scheme(line)
        return magic.retval

    @register_cell_magic
    def scheme(line, cell):
        magic.code = cell
        magic.cell_scheme()
        return magic.retval



# Copyright (c) Metakernel Development Team.
# Distributed under the terms of the Modified BSD License.

from metakernel import Magic, option
import os

class InstallMagic(Magic):

    def line_install(self, package):
        """
        %install PACKAGE - install package

        Example:
            %install calico-spell-check
        """
        ## FIXME: get list of known extension names and locations from wiki
        if package == "calico-publish":
            self.kernel.do_execute("!ipython install-nbextension https://bitbucket.org/ipre/calico/raw/master/notebooks/nbextensions/calico-publish.js")
        elif package == "calico-spell-check":
            self.kernel.do_execute("!ipython install-nbextension https://bitbucket.org/ipre/calico/downloads/calico-spell-check-1.0.zip")
        elif package == "calico-cell-tools":
            self.kernel.do_execute("!ipython install-nbextension https://bitbucket.org/ipre/calico/downloads/calico-cell-tools-1.0.zip")
        elif package == "calico-document-tools":
            self.kernel.do_execute("!ipython install-nbextension https://bitbucket.org/ipre/calico/downloads/calico-document-tools-1.0.zip")
        self.enable_extension(package)
        # FIXME: related %config:
        ## // To turn off automatically creating closing parenthesis and bracket:
        ## IPython.CodeCell.options_default.cm_config["autoCloseBrackets"] = "";

    def enable_extension(self, name):
        filename = "~/.ipython/profile_default/static/custom/custom.js"
        if filename.startswith("~"):
            filename = os.path.expanduser(filename)
        filename = os.path.abspath(filename)
        text = open(filename).read()
        if ('IPython.load_extensions("%s");' % name) in text:
            return
        if "// INSTALL MAGIC" not in text:
            text += """
require(["base/js/events"], function (events) {
    events.on("app_initialized.NotebookApp", function () {
        // INSTALL MAGIC
        // To turn off automatically creating closing parenthesis and bracket:
        IPython.CodeCell.options_default.cm_config["autoCloseBrackets"] = "";
    });
});
"""
        text = text.replace("        // INSTALL MAGIC", '        IPython.load_extensions("%s");\n        // INSTALL MAGIC' % name)
        with open(filename, "w") as fp:
            fp.write(text)

def register_magics(kernel):
    kernel.register_magics(InstallMagic)




from metakernel.tests.utils import get_kernel, get_log_text


def test_magic_magic():
    kernel = get_kernel()
    kernel.do_execute('%magic', None)
    text = get_log_text(kernel)
    assert '! COMMAND ... - execute command in shell' in text



import errno
import sys
import re
import signal
import os
import atexit

from . import pexpect

PY3 = (sys.version_info[0] >= 3)

if PY3:
    def u(s):
        return s
else:
    def u(s):
        if isinstance(s, str):
            return s.decode('utf-8')
        return s

PEXPECT_PROMPT = u('PEXPECT_PROMPT>')
PEXPECT_STDIN_PROMPT = u('PEXPECT_PROMPT+')
PEXPECT_CONTINUATION_PROMPT = u('PEXPECT_PROMPT_')


class REPLWrapper(object):

    """Wrapper for a REPL.

    All prompts are interpreted as regexes.  If you have special
    characters in the prompt, use `re.escape` to escape the characters.

    :param cmd_or_spawn: This can either be an instance of
    :class:`pexpect.spawn` in which a REPL has already been started,
    or a str command to start a new REPL process.
    :param str prompt_regex:  Regular expression representing process prompt, eg ">>>" in Python.
    :param str continuation_prompt_regex: Regular expression repesenting process continuation prompt, e.g. "..." in Python.
    :param str prompt_change_cmd: Optional kernel command that sets continuation-of-line-prompts, eg PS1 and PS2, such as "..." in Python.
        to something more unique. If this is ``None``, the prompt will not be
        changed. This will be formatted with the new and continuation prompts
        as positional parameters, so you can use ``{}`` style formatting to
        insert them into the command.
    :param str new_prompt_regex: The more unique prompt to expect after the change.
    :param str stdin_prompt_regex: The regex for a stdin prompt from the 
        child process.  The prompt itself will be sent to the `stdin_handler`,
        so any sentinel value inserted will have to be removed by the caller.
    :param str extra_init_cmd: Commands to do extra initialisation, such as
      disabling pagers.
    :param str prompt_emit_cmd: Optional kernel command that emits the prompt
      when one is not emitted by default (typically happens on Windows only)
    :param bool echo: Whether the child should echo, or in the case
    of Windows, whether the child does echo.
    """

    def __init__(self, cmd_or_spawn, prompt_regex, prompt_change_cmd,
                 new_prompt_regex=PEXPECT_PROMPT,
                 continuation_prompt_regex=PEXPECT_CONTINUATION_PROMPT,
                 stdin_prompt_regex=PEXPECT_STDIN_PROMPT,
                 extra_init_cmd=None,
                 prompt_emit_cmd=None,
                 echo=False):
        if isinstance(cmd_or_spawn, str):
            self.child = pexpect.spawnu(cmd_or_spawn, echo=echo,
                                        codec_errors="ignore",
                                        encoding="utf-8")
        else:
            self.child = cmd_or_spawn

        if self.child.echo and not echo:
            # Existing spawn instance has echo enabled, disable it
            # to prevent our input from being repeated to output.
            self.child.setecho(False)
            self.child.waitnoecho()

        # Convert all arguments to unicode.
        prompt_regex = u(prompt_regex)
        prompt_change_cmd = u(prompt_change_cmd)
        continuation_prompt_regex = u(continuation_prompt_regex)
        stdin_prompt_regex = u(stdin_prompt_regex)
        prompt_emit_cmd = u(prompt_emit_cmd)

        self.echo = echo
        self.prompt_emit_cmd = prompt_emit_cmd

        if prompt_change_cmd is None:
            self.prompt_regex = u(prompt_regex)
        else:
            self.set_prompt(prompt_regex,
                            prompt_change_cmd.format(new_prompt_regex,
                                                     continuation_prompt_regex))
            self.prompt_regex = new_prompt_regex
        self.continuation_prompt_regex = continuation_prompt_regex
        self.stdin_prompt_regex = stdin_prompt_regex

        self._stream_handler = None
        self._stdin_handler = None
        self._expect_prompt()

        if extra_init_cmd is not None:
            self.run_command(extra_init_cmd)

        atexit.register(self.terminate)

    def sendline(self, line):
        self.child.sendline(u(line))
        if self.echo:
            self.child.readline()

    def set_prompt(self, prompt_regex, prompt_change_cmd):
        self.child.expect(prompt_regex)
        self.sendline(prompt_change_cmd)

    def _expect_prompt(self, timeout=None):
        """Expect a prompt from the child.
        """
        stream_handler = self._stream_handler
        stdin_handler = self._stdin_handler
        expects = [self.prompt_regex, self.continuation_prompt_regex,
                   self.stdin_prompt_regex]
        if stream_handler:
            expects += [u(self.child.crlf)]
        if self.prompt_emit_cmd:
            self.sendline(self.prompt_emit_cmd)
        while True:
            pos = self.child.expect(expects, timeout=timeout)
            if pos == 2 and stdin_handler:
                resp = stdin_handler(self.child.before + self.child.after)
                self.sendline(resp)
            elif pos == 3:  # End of line received
                stream_handler(self.child.before)
            else:
                if len(self.child.before) != 0 and stream_handler:
                    # prompt received, but partial line precedes it
                    stream_handler(self.child.before)
                break
        return pos

    def run_command(self, command, timeout=None, stream_handler=None,
                    stdin_handler=None):
        """Send a command to the REPL, wait for and return output.
        :param str command: The command to send. Trailing newlines are not needed.
          This should be a complete block of input that will trigger execution;
          if a continuation prompt is found after sending input, :exc:`ValueError`
          will be raised.
        :param int timeout: How long to wait for the next prompt. -1 means the
          default from the :class:`pexpect.spawn` object (default 30 seconds).
          None means to wait indefinitely.
        """
        # Split up multiline commands and feed them in bit-by-bit
        cmdlines = command.splitlines()
        # splitlines ignores trailing newlines - add it back in manually
        if command.endswith('\n'):
            cmdlines.append('')
        if not cmdlines:
            raise ValueError("No command was given")

        res = []
        self._stream_handler = stream_handler
        self._stdin_handler = stdin_handler
        self.sendline(cmdlines[0])
        for line in cmdlines[1:]:
            self._expect_prompt(timeout=timeout)
            res.append(self.child.before)
            self.sendline(line)

        # Command was fully submitted, now wait for the next prompt
        if self._expect_prompt(timeout=timeout) == 1:
            # We got the continuation prompt - command was incomplete
            self.interrupt()
            raise ValueError("Continuation prompt found - input was incomplete:\n" + command)
        return u''.join(res + [self.child.before])

    def interrupt(self):
        """Interrupt the process and wait for a prompt.

        Returns
        -------
        The value up to the prompt.
        """
        if pexpect.pty:
            self.child.sendintr()
        else:
            self.child.kill(signal.SIGINT)
        while 1:
            try:
                self._expect_prompt(timeout=-1)
                break
            except KeyboardInterrupt:
                pass
        return self.child.before

    def terminate(self):
        if pexpect.pty:
            self.child.close()
            return self.child.terminate()
        try:
            self.child.kill(signal.SIGTERM)
        except Exception as e:
            if e.errno != errno.EACCES:
                raise


def python(command="python"):
    """Start a Python shell and return a :class:`REPLWrapper` object."""
    if not pexpect.pty:
        raise OSError('Not supported on platform "%s"' % sys.platform)
    return REPLWrapper(command, u(">>> "),
                       u("import sys; sys.ps1={0!r}; sys.ps2={1!r}"))


def bash(command="bash", prompt_regex=re.compile('[$#]')):
    """Start a bash shell and return a :class:`REPLWrapper` object."""

    # If the user runs 'env', the value of PS1 will be in the output. To avoid
    # replwrap seeing that as the next prompt, we'll embed the marker characters
    # for invisible characters in the prompt; these show up when inspecting the
    # environment variable, but not when bash displays the prompt.
    ps1 = PEXPECT_PROMPT[:5] + u'\[\]' + PEXPECT_PROMPT[5:]
    ps2 = PEXPECT_CONTINUATION_PROMPT[:5] + u'\[\]' + PEXPECT_CONTINUATION_PROMPT[5:]
    prompt_change_cmd = u"PS1='{0}' PS2='{1}' PROMPT_COMMAND=''".format(ps1, ps2)

    if os.name == 'nt':
        prompt_regex = u('__repl_ready__')
        prompt_emit_cmd = u('echo __repl_ready__')
        prompt_change_cmd = None
    else:
        prompt_emit_cmd = None

    extra_init_cmd = "export PAGER=cat"

    # Make sure the bash shell has a valid ending character.
    bashrc = os.path.join(os.path.dirname(pexpect.PEXPECT_DIR), 'bashrc.sh')
    child = pexpect.spawn(command, ['--rcfile', bashrc], echo=False,
                          encoding='utf-8')

    return REPLWrapper(child, prompt_regex, prompt_change_cmd,
                       prompt_emit_cmd=prompt_emit_cmd,
                       extra_init_cmd=extra_init_cmd)


def cmd(command='cmd', prompt_regex=re.compile(r'[A-Z]:\\.*>')):
    """"Start a cmd shell and return a :class:`REPLWrapper` object."""
    if not os.name == 'nt':
        raise OSError('cmd only available on Windows')
    return REPLWrapper(command, prompt_regex, None, echo=True)
